# =============================================================================
# Fog Graph Sparsification
# =============================================================================
#
# Sparsification schemes for networks.
#
import math
from fog.graph.utils import second_largest_component_size


def floatsam_sparsification(graph, starting_treshold=0.0, learning_rate=0.05,
                            max_drifter_size=None, weight='weight',
                            remove_edges=False):
    """
    Function using an iterative algorithm to try and find the best weight
    threshold to apply to trim the given graph's edges while keeping the
    underlying community structures.

    It works by iteratively increasing the threshold and stopping as soon as
    a significant connected component starts to drift away from the principal
    one.

    This is basically a very naive gradient descent with a very naive cost
    function but it works decently for typical cases.

    Args:
        graph (nx.Graph): Graph to sparsify.
        starting_treshold (float, optional): Starting similarity threshold.
            Defaults to `0.0`.
        learning_rate (float, optional): How much to increase the threshold
            at each step of the algorithm. Defaults to `0.05`.
        max_drifter_size (int, optional): Max size of component to detach itself
            from the principal one before stopping the algorithm. If not
            provided it will default to the logarithm of the graph's total
            number of nodes.
        weight (str, optional): Name of the weight attribute.
            Defaults to "weight" wrt networkx conventions.
        remove_edges (bool, optional): Whether to remove edges from the graph
            having a weight less than found threshold or not. Note that if
            `True`, this will mutate the given graph. Defaults to `False`.

    Returns:
        float: The found threshold

    """
    if graph.size() == 0:
        return starting_treshold

    threshold = starting_treshold
    best_threshold = None

    def edge_filter(u, v, attr):
        return attr[weight] >= threshold

    if max_drifter_size is None:
        max_drifter_size = int(math.log(len(graph)))

    while True:
        best_threshold = threshold
        threshold += learning_rate

        c = second_largest_component_size(graph, edge_filter)

        if c is not None and c >= max_drifter_size:
            break

    if remove_edges:
        to_remove = []
        for u, v, w in graph.edges(data=weight):
            if w < best_threshold:
                to_remove.append((u, v))

        for u, v in to_remove:
            graph.remove_edge(u, v)

    return best_threshold



# =============================================================================
# Fog Fingerprint Tokenizers
# =============================================================================
#
# Functions for creating custom fingerprint tokenizers.
#
# Fingerprinting is a clever way to normalize strings in order to remove
# anything that would not compulsorily yield meaning such as token ordering,
# & duplication etc.
#
# The basic fingerprinting procedure is the following:
#   1. Normalizing the case
#   2. Normalizing unicode characters to plain ascii
#   3. Dropping any non-alphanumeric character
#   4. Dropping irrelevant whitespace
#   5. Tokenizing by splitting the string by whitespaces
#   6. Deduplicating the tokens
#   7. Sorting the tokens
#   8. Returning those tokens that will usually be re-joined by a single space
#
# For instance, this fingerprinting scheme will be able to match the
# following strings:
#   "University of North Carolina"
#   "North Carolina, university of"
#   "UNIVERSITY OF NORTH CAROLINA"
#   "University   of of north CarolINA"
#   "University --- of --- North Carolina"
#   "..."
#
# Optionally, one can chose more advanced and aggressive options such as
# filtering some tokens and squeezing the string etc.
#
# The ngram fingerprint is a variant that returns a list of normalized ngrams
# that will be joined in order rather than the whole tokens. It is
# sometimes useful to find misspelled strings.
#
# [Url]:
# http://openrefine.org/
#
import re
from unidecode import unidecode
from fog.utils import squeeze as squeeze_string
from fog.tokenizers.ngrams import ngrams

WHITESPACE_RE = re.compile('\\s+')
DIGITS_RE = re.compile('\\d+')
BLACKLIST_RE = re.compile('[^a-z0-9\\s]')


class FingerprintTokenizer(object):
    """
    Function returning a custom fingerprint tokenizer.

    Args:
        keep_digits (bool, optional): Whether to keep digits in fingerprint?
            Defaults to True.
        min_token_size (int, optional): Minimum token size. Defaults to 1.
        split (list, optional): List of token splitting characters.
        stopwords (iterable, optional): List of stopwords to filter.
        squeeze (bool, optional): Whether to squeeze the tokens or not.
    """

    def __init__(self, keep_digits=True, min_token_size=1,
                 split=None, stopwords=None, squeeze=False):

        self.keep_digits = keep_digits
        self.squeeze = squeeze

        self.stopwords_re = None
        self.split_re = None
        self.min_token_size_re = None

        if stopwords is not None:
            self.stopwords_re = re.compile(
                '(?:' + '|'.join(['\\b' + re.escape(s) + '\\b' for s in stopwords]) + ')',
                re.I
            )

        if split is not None:
            self.split_re = re.compile('[' + re.escape(''.join(split)) + ']')

        if min_token_size and min_token_size > 1:
            self.min_token_size_re = re.compile('\\b\\S{1,%i}\\b' % min_token_size)

    def tokenize(self, string):

        # Splitting
        if self.split_re is not None:
            string = re.sub(self.split_re, ' ', string)

        # Stopwords
        if self.stopwords_re is not None:
            string = re.sub(self.stopwords_re, '', string)

        # Digits
        if not self.keep_digits:
            string = re.sub(DIGITS_RE, '', string)

        # Case normalization
        string = string.lower()

        # Minimum token size
        if self.min_token_size_re:
            string = re.sub(self.min_token_size_re, '', string)

        # Deburring
        string = unidecode(string)

        # Only keeping letters
        string = re.sub(BLACKLIST_RE, '', string)

        # Trimming
        string = string.strip()

        # Squeezing
        if self.squeeze:
            string = squeeze_string(string)

        # Tokenizing
        tokens = re.split(WHITESPACE_RE, string)

        # Keeping unique tokens
        tokens = set(tokens)

        # Sorting
        tokens = sorted(tokens)

        return tokens

    def __call__(self, string):
        return self.tokenize(string)

    def key(self, string):
        return ' '.join(self.tokenize(string))


class NgramsFingerprintTokenizer(FingerprintTokenizer):
    def tokenize(self, n, string):

        # Splitting
        if self.split_re is not None:
            string = re.sub(self.split_re, ' ', string)

        # Stopwords
        if self.stopwords_re is not None:
            string = re.sub(self.stopwords_re, '', string)

        # Digits
        if not self.keep_digits:
            string = re.sub(DIGITS_RE, '', string)

        # Case normalization
        string = string.lower()

        # Minimum token size
        if self.min_token_size_re:
            string = re.sub(self.min_token_size_re, '', string)

        # Deburring
        string = unidecode(string)

        # Only keeping letters
        string = re.sub(BLACKLIST_RE, '', string)

        # Trimming
        string = string.strip()

        # Squeezing
        if self.squeeze:
            string = squeeze_string(string)

        # Tokenizing
        string = re.sub(WHITESPACE_RE, '', string)
        tokens = ngrams(n, string)

        # Keeping unique tokens
        tokens = set(tokens)

        # Sorting
        tokens = sorted(tokens)

        return tokens

    def __call__(self, n, string):
        return self.tokenize(n, string)

    def key(self, n, string):
        return ''.join(self.tokenize(n, string))



from fog.clustering.blocking import blocking
from fog.clustering.intersection_index import (
    intersection_index
)
from fog.clustering.key_collision import key_collision
from fog.clustering.laesa import laesa
from fog.clustering.minhash import minhash
from fog.clustering.nn_descent import nn_descent, nn_descent_full
from fog.clustering.pairwise import (
    pairwise,
    pairwise_leader,
    pairwise_fuzzy_clusters,
    pairwise_connected_components
)
from fog.clustering.passjoin import passjoin
from fog.clustering.ppjoin import (
    all_pairs,
    ppjoin,
    ppjoin_plus
)
from fog.clustering.quickjoin import quickjoin
from fog.clustering.sorted_neighborhood import (
    sorted_neighborhood,
    adaptive_sorted_neighborhood
)
from fog.clustering.vp_tree import vp_tree



from setuptools import setup, find_packages, Extension

with open('./README.md', 'r') as f:
    long_description = f.read()

EXTENSIONS = [
    Extension(
        'cfog.metrics.levenshtein',
        ['cfog/metrics/levenshtein.c'],
        optional=True
    )
]

setup(name='fog',
    version='0.11.9',
    description='A fuzzy matching & clustering library for python.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='http://github.com/Yomguithereal/fog',
    license='MIT',
    author='Guillaume Plique',
    author_email='kropotkinepiotr@gmail.com',
    keywords='fuzzy',
    python_requires='>=3',
    packages=find_packages(exclude=['experiments', 'experiments.*', 'test', 'test.*']),
    ext_modules=EXTENSIONS,
    install_requires=[
        'dill>=0.2.7.1',
        'ebbe>=1.3.1,<2',
        'emoji>=1,<2',
        'networkx>=2',
        'phylactery>=0.1.1',
        'Unidecode>=1.0.22'
    ],
    entry_points={
        'console_scripts': ['fog=fog.cli.__main__:main']
    },
    zip_safe=True)



# =============================================================================
# Fog MinHash
# =============================================================================
#
# Classes & functions related to the MinHash Local Sensitivity Hashing scheme.
#
# [Urls]:
# https://en.wikipedia.org/wiki/MinHash
# https://arxiv.org/abs/1706.05698
#
# [Reference]:
# Broder, Andrei Z. (1997), "On the resemblance and containment of documents",
# Compression and Complexity of Sequences: Proceedings, Positano,
# Amalfitan Coast, Salerno, Italy, June 11-13, 1997.
#
# Ertl, Otmar. « SuperMinHash - A New Minwise Hashing Algorithm for Jaccard
# Similarity Estimation ». arXiv:1706.05698 [cs], 18 juin 2017.
#
import math
from random import Random

try:
    import numpy as np
except:
    np = None

from fog.lsh.utils import crc32, popcount64

MAX_UINT32 = (2 ** 32) - 1
NEXT_PRIME = 4294967311


def minhash_similarity(A, B):
    hamming = 0
    h = len(A)

    for i in range(h):
        if A[i] != B[i]:
            hamming += 1

    return 1.0 - hamming / h


def lsb_minhash_similarity(A, B):
    hamming = 0
    L = len(A)
    h = L * 64.0

    for i in range(L):
        hamming += popcount64(A[i] ^ B[i])

    return 1.0 - 2.0 * hamming / h


class MinHash(object):

    __slots__ = ('A', 'B', 'params', 'h', 'numpy')

    def __init__(self, h=256, seed=None, use_numpy=False):

        # Properties
        self.A = None
        self.B = None
        self.params = None

        if use_numpy:
            assert np is not None, 'numpy is not installed'

            np.random.seed(seed)

            self.A = np.random.randint(1, MAX_UINT32, size=h, dtype='uint32')
            self.B = np.random.randint(0, MAX_UINT32, size=h, dtype='uint32')

            self.numpy = True
        else:

            rng = Random(seed)

            params = set()

            while len(params) < h:
                params.add((
                    rng.randint(1, MAX_UINT32),
                    rng.randint(0, MAX_UINT32)
                ))

            self.params = list(params)
            self.numpy = False

        self.h = h

    def create_signature(self, sequence):

        if type(sequence) is str:
            tokens = set(ord(c) for c in sequence)
        else:
            tokens = set(crc32(token) for token in sequence)

        # Using numpy
        if self.numpy:

            if len(tokens) == 0:
                return np.zeros(self.h, dtype=np.uint32)

            signature = np.repeat(list(tokens), self.h).astype(np.uint32, copy=False)
            signature.shape = (len(tokens), self.h)

            # Universal hash
            signature *= self.A
            signature += self.B
            signature %= NEXT_PRIME

            return signature.min(axis=0)

        # Standard routine
        params = self.params
        signature = [0] * self.h

        if len(tokens) == 0:
            return signature

        for s in range(self.h):
            min_hash = MAX_UINT32
            a, b = params[s]

            for token in tokens:
                h = (a * token + b) % NEXT_PRIME

                if h < min_hash:
                    min_hash = h

            signature[s] = min_hash

        return signature


class SuperMinHash(object):

    def __init__(self, h=256):
        self.h = h
        self.rng = Random()

    def create_signature(self, sequence):

        if type(sequence) is str:
            tokens = set(ord(c) for c in sequence)
        else:
            tokens = set(crc32(token) for token in sequence)

        m = self.h
        rng = self.rng

        h = [m + 1] * m
        p = [0] * m
        q = [-1] * m
        b = [0] * m

        b[-1] = m

        a = m - 1

        for i, token in enumerate(tokens):
            rng.seed(token)
            j = 0

            while j <= a:
                r = rng.random()
                k = j + math.floor(r * (m - j))

                if q[j] != i:
                    q[j] = i
                    p[j] = j

                if q[k] != i:
                    q[k] = i
                    p[k] = k

                tmp = p[j]
                p[j] = p[k]
                p[k] = tmp

                if r + j < h[p[j]]:
                    j2 = min(math.floor(h[p[j]]), m - 1)
                    h[p[j]] = r + j

                    if j < j2:
                        b[j2] -= 1
                        b[j] += 1

                        while b[a] == 0:
                            a -= 1

                j += 1

        return [int(i) for i in h]


class LSBMinHash(object):

    def __init__(self, precision=8, seed=None):
        # TODO: cheap_hashes

        rng = Random(seed)

        h = precision * 64

        params = set()

        while len(params) < h:
            params.add((
                rng.randint(1, MAX_UINT32),
                rng.randint(0, MAX_UINT32)
            ))

        self.params = list(params)
        self.precision = precision

    def create_signature(self, sequence):
        params = self.params

        if type(sequence) is str:
            tokens = set(ord(c) for c in sequence)
        else:
            tokens = set(crc32(token) for token in sequence)

        signature = [0] * self.precision

        if len(tokens) == 0:
            return signature

        # TODO: numpy?
        for s in range(self.precision):
            integer = 0
            offset = s * 64

            for i in range(64):
                min_hash = MAX_UINT32
                a, b = params[offset + i]

                for token in tokens:
                    h = (a * token + b) % NEXT_PRIME

                    if h < min_hash:
                        min_hash = h

                integer |= ((min_hash & 0x1) << i)

            signature[s] = integer

        return signature



from fog.metrics.cosine import (
    cosine_similarity,
    sparse_cosine_similarity,
    sparse_dot_product,
    sparse_norm,
    sparse_normalize,
    binary_cosine_similarity,
    sparse_binary_cosine_similarity,
)
from fog.metrics.dice import (
    dice_coefficient
)
from fog.metrics.jaccard import (
    jaccard_similarity,
    weighted_jaccard_similarity
)
from fog.metrics.overlap import overlap_coefficient



# =============================================================================
# Fog Blocking Clustering
# =============================================================================
#
# Implementation of a blocking clusterer that starts by dispatching given
# items to one or more buckets before computing pairwise comparisons on them.
#
import dill
from collections import defaultdict
from multiprocessing import Pool
from fog.clustering.utils import make_similarity_function, clusters_from_pairs

# TODO: max_block_size to avoid ngrams with high DF
# TODO: worker using a VPTree
# TODO: custom inner algorithm
# TODO: fuzzy blocking variant
# TODO: possibility not to merge if sure cannot collide twice
# TODO: the worker_graph is not used right now, the code does not use it (make a clusters from graph also)


def block_worker(payload):
    """
    Worker function used compute pairwise distance/similarity over a whole
    block.

    """
    similarity, block, serialized, graph = payload

    if serialized:
        similarity = dill.loads(similarity)

    pairs = []
    n = len(block)

    for i in range(n):
        A = block[i]

        for j in range(i + 1, n):
            B = block[j]

            if graph is not None and A in graph and B in graph[A]:
                continue

            if similarity(A, B):
                pairs.append((A, B))

    return pairs


def blocking(data, block=None, blocks=None, similarity=None, distance=None,
             radius=None, min_size=2, max_size=float('inf'), processes=1,
             mode='connected_components'):
    """
    Function returning an iterator over found clusters using the blocking
    method.

    It works by dispatching given items into one or more buckets before
    computing pairwise comparisons on each bucket.

    Args:
        data (iterable): Arbitrary iterable containing data points to gather
            into clusters. Will be fully consumed.
        block (callable): A function returning an item's block.
        blocks (callable): A function returning an item's blocks.
        similarity (callable): If radius is specified, a function returning
            the similarity between two points. Else, a function returning
            whether two points should be deemed similar. Alternatively, one can
            specify `distance` instead.
        distance (callable): If radius is specified, a function returning
            the distance between two points. Else, a function returning
            whether two point should not be deemed similar. Alternatively, one
            can specify `similarity` instead.
        radius (number, optional): produced clusters' radius.
        min_size (number, optional): minimum number of items in a cluster for
            it to be considered viable. Defaults to 2.
        max_size (number, optional): maximum number of items in a cluster for
            it to be considered viable. Defaults to infinity.
        processes (number, optional): number of processes to use.
            Defaults to 1.
        mode (string, optional): 'fuzzy_clusters', 'connected_components'.
            Defaults to 'connected_components'.

    Yields:
        list: A viable cluster.

    """

    # Formatting similarity
    similarity = make_similarity_function(similarity=similarity, distance=distance, radius=radius)

    # Single block or blocks?
    if blocks is None:
        graph = defaultdict(list)
        worker_graph = None

        def add(x, y):
            x.append(y)
    else:
        graph = defaultdict(set)
        worker_graph = graph

        def add(x, y):
            x.add(y)

    # Grouping items into buckets
    buckets = defaultdict(list)

    for item in data:
        if blocks is None:
            buckets[block(item)].append(item)
        else:
            bs = set(blocks(item))

            for b in bs:
                buckets[b].append(item)

    # Actual clustering
    def clustering():
        if processes == 1:
            for bucket in buckets.values():
                if len(bucket) < 2:
                    continue

                if type(bucket) is not list:
                    bucket = list(bucket)

                yield from block_worker((similarity, bucket, False, worker_graph))
        else:

            pickled_similarity = dill.dumps(similarity)

            pool_iter = (
                (pickled_similarity, bucket if type(bucket) is list else list(bucket), True, None)
                for bucket
                in buckets.values()
                if len(bucket) > 1
            )

            with Pool(processes=processes) as pool:
                for pairs in pool.imap_unordered(block_worker, pool_iter):
                    yield from pairs

    yield from clusters_from_pairs(
        clustering(),
        fuzzy=blocks is not None,
        min_size=min_size,
        max_size=max_size,
        mode=mode
    )



# =============================================================================
# Fog Jaccard Similarity
# =============================================================================
#
# Functions computing the Jaccard similarity
#
# [Urls]:
# https://en.wikipedia.org/wiki/Jaccard_index
# https://pdfs.semanticscholar.org/74be/e1ebf204dba4b2da0399a25a5ac9253a824e.pdf
#
# [References]:
# Jaccard, Paul (1912), “The distribution of the flora in the alpine zone”,
# New Phytologist 11: 37–50
#
# Ioffe, "Improved Consistent Sampling, Weighted Minhash and L1 Sketching".
#
from fog.metrics.utils import ACCEPTABLE_TYPES


def jaccard_similarity(A, B):
    """
    Function computing the Jaccard similarity. That is to say the intersection
    of input sets divided by their union.

    Runs in O(n), n being the size of the smallest set.

    Args:
        A (iterable): First sequence.
        B (iterable): Second sequence.

    Returns:
        float: Jaccard similarity between A & B.

    Example:
        from fog.metrics import jaccard_similarity

        # Basic
        jaccard_similarity('context', 'contact')
        >>> ~0.571

    """
    if A is B:
        return 1.0

    if not isinstance(A, ACCEPTABLE_TYPES):
        A = set(A)

    if not isinstance(B, ACCEPTABLE_TYPES):
        B = set(B)

    if len(A) == 0 or len(B) == 0:
        return 0.0

    # Swapping to iterate over smaller set and minimize lookups
    if len(A) > len(B):
        A, B = B, A

    # Counting intersection
    I = 0

    for v in A:
        if v in B:
            I += 1

    # Size of the union is size of A + size of B - intersection
    U = len(A) + len(B) - I

    return I / U


def weighted_jaccard_similarity(A, B):
    """
    Function computing the weighted Jaccard similarity.

    Runs in O(n), n being the sum of A & B's sizes.

    Args:
        A (Counter): First weighted set.
        B (Counter): Second weighted set.

    Returns:
        float: Weighted Jaccard similarity between A & B.

    Example:
        from fog.metrics import weighted_jaccard_similarity

        # Basic
        weighted_jaccard_similarity({'apple': 34, 'pear': 3}, {'pear': 1, 'orange': 1})
        >>> ~0.026

    """

    # Early termination
    if A is B:
        return 1.0

    if len(A) == 0 or len(B) == 0:
        return 0.0

    I = 0.0
    U = 0.0

    done = set()

    # Swapping to iterate over smaller set and minimize lookups
    if len(A) > len(B):
        A, B = B, A

    # Computing intersection
    for k, weight_A in A.items():
        weight_B = B.get(k, 0.0)

        if weight_B != 0.0:
            done.add(k)

        if weight_A < weight_B:
            I += weight_A
            U += weight_B
        else:
            I += weight_B
            U += weight_A

    # Finalizing union
    for k, v in B.items():
        if k in done:
            continue

        U += v

    return I / U



# =============================================================================
# Fog Fingerprint
# =============================================================================
#
# Functions returning string fingerprints.
#
from fog.tokenizers.fingerprint import (
    FingerprintTokenizer,
    NgramsFingerprintTokenizer
)


class FingerprintingKeyer(FingerprintTokenizer):
    def __call__(self, string):
        return self.key(string)


class NgramsFingerprintKeyer(NgramsFingerprintTokenizer):
    def __call__(self, n, string):
        return self.key(n, string)



# =============================================================================
# Fog Fagin Clustering
# =============================================================================
#
# Miscellaneous functions related to Fagin's algorithms in order to perform
# top k queries and such.
#
# [Articles]:
# Optimal aggregation algorithms for middleware
# Ronald Fagin, Amnon Lotem, and Moni Naor
#
# Combining Fuzzy Information from Multiple Systems
# Ronald Fagin
#
# Index-based, High-dimensional, Cosine Threshold Querying with Optimality
# Guarantees. Yuliang Li, Jianguo Wang, Benjamin Pullman, Nuno Bandeira, and
# Yannis Papakonstantinou
#
# [References]:
# http://homepages.inf.ed.ac.uk/libkin/teach/dataintegr09/topk.pdf
# http://alumni.cs.ucr.edu/~skulhari/Top-k-Query.pdf
# https://arxiv.org/pdf/1011.2807.pdf
# https://nlp.stanford.edu/IR-book/html/htmledition/contents-1.html
#
import math
from random import shuffle
from collections import defaultdict, Counter

from fog.metrics.cosine import sparse_cosine_similarity


def fagin_k1(vectors):
    inverted_lists = defaultdict(list)

    for i, vector in enumerate(vectors):
        for d, w in vector.items():
            inverted_lists[d].append((w, i))

    for l in inverted_lists.values():
        l.sort()

    for i, vector in enumerate(vectors):
        counts = Counter()
        offset = 0

        D = len(vector)

        while True:
            stop = True

            for d in vector:
                l = inverted_lists[d]

                if offset < len(l):
                    counts[l[offset][1]] += 1
                    stop = False

            if stop:
                break

            if sum(1 for c in counts.values() if c >= D) >= 2:
                yield i, list(j for j in counts.keys() if j != i)
                break

            offset += 1


def threshold_algorithm_k1(vectors):
    inverted_lists = defaultdict(list)

    for i, vector in enumerate(vectors):
        for d, w in vector.items():
            inverted_lists[d].append((w, i))

    for l in inverted_lists.values():
        l.sort()

    for i, vector in enumerate(vectors):
        visited = set()
        offset = 0

        t = 0.0
        best = None
        t_vector = {}

        while True:
            stop = True

            for d in vector:
                l = inverted_lists[d]

                if offset >= len(l):
                    continue

                stop = False

                w, j = l[offset]

                if i == j:

                    # TODO: probably ned to reset stop to take edge case into account
                    offset2 = offset + 1

                    if offset2 >= len(l):
                        continue

                    w, j = l[offset2]

                t_vector[d] = w

                if j in visited:
                    continue

                cs = sparse_cosine_similarity(vector, vectors[j])
                visited.add(j)

                if best is None or cs > best[0]:
                    best = (cs, j)

            # Final break + return self if best cos is 0.0
            if stop:
                yield i, best[1] if best is not None else i
                break

            t = sparse_cosine_similarity(vector, t_vector)

            if best is not None and best[0] >= t:
                yield i, best[1] if best is not None else i
                break

            offset += 1


def naive_cosine_pairs(vectors):
    inverted_lists = defaultdict(list)

    for i, vector in enumerate(vectors):
        visited = set()

        for d in vector:
            l = inverted_lists[d]

            for j in l:
                if j in visited:
                    continue

                visited.add(j)

                yield i, j

            l.append(i)


def sqrt_indexation_pairs(vectors):
    indices = list(range(len(vectors)))
    shuffle(indices)

    leaders_count = int(math.sqrt(len(vectors)))

    leaders = indices[:leaders_count]

    proximities = defaultdict(list)

    for i, v in enumerate(vectors):
        leader = min(leaders, key=lambda x: 1.0 - sparse_cosine_similarity(v, vectors[x]))

        yield i, leader

        l = proximities[leader]

        for j in l:
            yield i, j

        l.append(i)



from fog.lsh.minhash import (
    MinHash,
    LSBMinHash,
    SuperMinHash,
    minhash_similarity,
    lsb_minhash_similarity
)
from fog.lsh.simhash import (
    simhash,
    simhash_distance,
    simhash_similarity
)



# =============================================================================
# Fog Cosine Similarity
# =============================================================================
#
# Functions computing the cosine similarity.
#
# [Url]:
# https://en.wikipedia.org/wiki/Cosine_similarity
#
import math
from collections import Counter

from fog.metrics.utils import ACCEPTABLE_TYPES

COSINE_ACCEPTABLE_TYPES = (dict, Counter)


def sparse_cosine_similarity(A, B):
    """
    Function computing cosine similarity on sparse weighted sets represented
    as python dicts.

    Runs in O(n), n being the sum of A & B's sizes.

    Args:
        A (Counter): First weighted set.
        B (Counter): Second weighted set.

    Returns:
        float: Cosine similarity between A & B.

    Example:
        from fog.metrics import sparse_cosine_similarity

        # Basic
        sparse_cosine_similarity({'apple': 34, 'pear': 3}, {'pear': 1, 'orange': 1})
        >>> ~0.062

    """

    # Early termination
    if A is B:
        return 1.0

    if len(A) == 0 or len(B) == 0:
        return 0.0

    xx = 0.0
    xy = 0.0
    yy = 0.0

    # Swapping to iterate over smaller set and minimize lookups
    if len(A) > len(B):
        A, B = B, A

    for k, v in A.items():
        weight = v
        xx += weight ** 2

        v2 = B.get(k)

        if v2 is not None:
            xy += weight * v2

    for v in B.values():
        weight = v
        yy += weight ** 2

    return xy / math.sqrt(xx * yy)


def sparse_binary_cosine_similarity(A, B):
    """
    Function computing binary cosine similarity on sparse vectors represented
    as python sets.

    Runs in O(n), n being the size of the smaller set.

    Args:
        A (Counter): First set.
        B (Counter): Second set.

    Returns:
        float: Binary cosine similarity between A & B.

    """

    # Early termination
    if A is B:
        return 1.0

    if len(A) == 0 or len(B) == 0:
        return 0.0

    xx = 0.0
    xy = 0.0
    yy = 0.0

    # Swapping to iterate over smaller set and minimize lookups
    if len(A) > len(B):
        A, B = B, A

    for k in A:
        xx += 1

        if k in B:
            xy += 1

    yy += len(B)

    return xy / math.sqrt(xx * yy)


def sparse_dot_product(A, B):
    """
    Function used to compute the dotproduct of sparse weighted sets represented
    by python dicts.

    Runs in O(n), n being the size of the smallest set.

    Args:
        A (Counter): First weighted set.
        B (Counter): Second weighted set.

    Returns:
        float: Dot product of A & B.

    """

    # Swapping so we iterate over the smallest set
    if len(A) > len(B):
        A, B = B, A

    product = 0.0

    for k, w1 in A.items():
        w2 = B.get(k)

        if w2 is not None:
            product += w1 * w2

    return product


def sparse_norm(A):
    return math.sqrt(sum(w * w for w in A.values()))


def sparse_normalize(A):
    N = {}

    length = sparse_norm(A)

    if length == 0:
        return N

    for k, w in A.items():
        N[k] = w / length

    return N


def cosine_similarity(A, B):
    """
    Function computing the cosine similarity of the given sequences.

    Runs in O(n), n being the sum of A & B's sizes.

    Args:
        A (iterable): First sequence.
        B (iterable): Second sequence.

    Returns:
        float: Cosine similarity between A & B.

    """

    # Computing frequencies
    if not isinstance(A, COSINE_ACCEPTABLE_TYPES):
        A = Counter(A)

    if not isinstance(B, COSINE_ACCEPTABLE_TYPES):
        B = Counter(B)

    return sparse_cosine_similarity(A, B)


def binary_cosine_similarity(A, B):
    """
    Function computing the binary cosine similarity of the given sequences.

    Runs in O(n), n being the size of the smallest set.

    Args:
        A (iterable): First sequence.
        B (iterable): Second sequence.

    Returns:
        float: Binary cosine similarity between A & B.

    """

    # Computing frequencies
    if not isinstance(A, ACCEPTABLE_TYPES):
        A = set(A)

    if not isinstance(B, ACCEPTABLE_TYPES):
        B = set(B)

    return sparse_binary_cosine_similarity(A, B)



# =============================================================================
# Fog Omission Key
# =============================================================================
#
# The omission key by Pollock and Zamora.
#
# [Urls]:
# http://dl.acm.org/citation.cfm?id=358048
# http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.12.385&rep=rep1&type=pdf
#
# [Reference]:
# Pollock, Joseph J. and Antonio Zamora. 1984. "Automatic Spelling Correction
# in Scientific and Scholarly Text." Communications of the ACM, 27(4).
# 358--368.
import re
from unidecode import unidecode

UNDESIRABLES_RE = re.compile(r'[^A-Z]')
CONSONANTS = 'JKQXZVWYBFMGPDHCLNTSR'
VOWELS = set('AEIOU')


def omission_key(string):
    """
    Function returning a string's omission key which is constructed thusly:

    1. First we record the string's set of consonant in an order
       where most frequently mispelled consonants will be last.
    2. Then we record the string's set of vowels in the order of
       first appearance.

    This key is very useful when searching for mispelled strings because
    if sorted using this key, similar strings will be next to each other.

    Args:
        string (str): The string to encode.

    Returns:
        string: The string's omission key.

    """

    # Deburring
    string = unidecode(string)

    # Normalizing case
    string = string.upper()

    # Dropping useless characters
    string = re.sub(UNDESIRABLES_RE, '', string)

    if not string:
        return ''

    # Composing the key
    letters = set()
    consonants = []
    vowels = []

    # Adding vowels in order they appeared
    for letter in string:
        if letter in VOWELS and letter not in vowels:
            vowels.append(letter)
        else:
            letters.add(letter)

    # Adding consonants in order
    for consonant in CONSONANTS:
        if consonant in letters:
            consonants.append(consonant)

    return ''.join(consonants + vowels)



# =============================================================================
# Fog Metrics Utilities
# =============================================================================
#
# Miscellaneous utility functions used to compute metrics
#
ACCEPTABLE_TYPES = (set, frozenset, dict)


def intersection_size(A, B):
    if A is B:
        return len(A)

    if not isinstance(A, ACCEPTABLE_TYPES):
        A = set(A)

    if not isinstance(B, ACCEPTABLE_TYPES):
        B = set(B)

    if len(A) > len(B):
        A, B = B, A

    if len(A) == 0:
        return 0

    i = 0

    for x in A:
        if x in B:
            i += 1

    return i
